﻿using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Transactions;
using System.Web;
using GestorDespeses.Exceptions;
using GestorDespeses.Utils;
using GestorDespeses.Model;

namespace GestorDespeses.BI
{
    public class CoreBI
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private GestorDespesesContext db;
        private bool initializedDB;

        public CoreBI()
        {
            log.Info("Initializing database context...");
            db = new GestorDespesesContext();
            initializedDB = true;
            log.Info("Context initialized");
        }

        public CoreBI(GestorDespesesContext context)
        {
            db = context;
            initializedDB = false;
        }

        ~CoreBI()
        {
            if (initializedDB)
            {
                log.Info("Disposing database context...");
                db.Dispose();
                log.Info("Context disposed");
            }
        }

        /// <summary>
        /// Envia un correu electronic al usuari que se li pasa com a paràmetre, utilitzant la configuració que hi ha al sistema
        /// </summary>
        /// <param name="user">Usuari</param>
        /// <param name="subject">Asumpte del correu</param>
        /// <param name="content">Contingut del correu</param>
        /// <returns>True si s'ha pogut enviar el correu, False si no</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool sendSMTP(User user, string subject, string content) 
        {
            log.Debug("sendSMTP Begin");
            bool result = false;
            try
            {
                log.Info("Sending email to " + user.Email + " with subject " + subject + ".");
                SMTPUtils.sendMail(getSetting(Constants.SMTP_FROM_KEY),
                    getSetting(Constants.SMTP_USERNAME_KEY),
                    getSetting(Constants.SMTP_PASSWORD_KEY),
                    getSetting(Constants.SMTP_HOST_KEY),
                    Convert.ToInt32(getSetting(Constants.SMTP_PORT_KEY)),
                    Convert.ToBoolean(getSetting(Constants.SMTP_SSL_KEY)),
                    user.Email, 
                    user.FullName, 
                    subject, 
                    content);
                result = true;
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al enviar un email.", e);
            }
            log.Debug("sendSMTP End");

            return result;
        }

        /// <summary>
        /// Canvia la configuració del servidor de correu del sistema
        /// </summary>
        /// <param name="fromEmail">Direcció de correu del remitent</param>
        /// <param name="fromFullName">Nom del remitent</param>
        /// <param name="password">Contrasenya del remitent per el host</param>
        /// <param name="host">Direcció del servidor de correu</param>
        /// <param name="port">Port del servidor de correu</param>
        /// <param name="enableSSL">True per habilitar SSL, False per enviaro en clar.</param>
        /// <returns>True si s'ha pogut modificar les dades, False sino.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool editSMTPSettings(string fromEmail,
            string fromFullName,
            string password,
            string host,
            int port,
            bool enableSsl)
        {
            log.Debug("editSMTPSetting Begin");
            bool result = false;
            try
            {
                using (TransactionScope transaction = new TransactionScope())
                {
                    if (!SMTPUtils.IsValidEmail(fromEmail))
                    {
                        throw new GestorDespesesException("Direcció de correu no vàlida.");
                    }
                    setSetting(Constants.SMTP_FROM_KEY, fromEmail);
                    setSetting(Constants.SMTP_USERNAME_KEY, fromFullName);
                    setSetting(Constants.SMTP_PASSWORD_KEY, password);
                    setSetting(Constants.SMTP_HOST_KEY, host);
                    setSetting(Constants.SMTP_PORT_KEY, Convert.ToString(port));
                    setSetting(Constants.SMTP_SSL_KEY, Convert.ToString(enableSsl));
                    transaction.Complete();
                }
                result = true;
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al establir els parametres de configuració SMTP.", e);
            }
            log.Debug("sendSMTP End");

            return result;
        }

        /// <summary>
        /// Crear o modifica un setting dins de la tabla de propietats del sistema
        /// </summary>
        /// <param name="key">Clau de la propietat</param>
        /// <param name="value">Valor de la propietat</param>
        /// <returns>True si s'ha pogut afegir/modificar, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool setSetting(string key, string value)
        {
            log.Debug("setSetting Begin");
            bool result = false;
            try
            {
                var properties = (from s in db.SystemProperties
                                           where s.Key == key
                                           select s);
                if (properties.Count() == 1)
                {
                    log.Info("Setting " + key + " property to " + value);
                    SystemProperty property = properties.First();
                    property.Value = value;
                }
                else
                {
                    log.Info("Creating " + key + " property with value " + value);
                    SystemProperty property = new SystemProperty();
                    property.Key = key;
                    property.Value = value;
                    db.SystemProperties.Add(property);
                }
                db.SaveChanges();
                result = true;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al introduir o editar un setting nou.", e);
            }
            log.Debug("setSetting End");

            return result; 
        }

        /// <summary>
        /// Recupera el valor de la propietat del sistema amb la clau que se li pasa com a paràmetre
        /// </summary>
        /// <param name="key">Clau de la propietat</param>
        /// <returns>Valor de la propietat.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public string getSetting(string key)
        {
            log.Debug("getSetting Begin");
            string result = null;
            try
            {
                var properties = (from s in db.SystemProperties
                                  where s.Key == key
                                  select s);
                if (properties.Count() != 1)
                {
                    throw new GestorDespesesException("No s'ha trobat cap propietat del sistema amb clau " + key);
                }
                result = properties.First().Value;
                log.Info("Accessing property " + key + " with value " + result);
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar una propietat del sistema.", e);
            }
            log.Debug("getSetting End");
            return result; 
        }

        /// <summary>
        /// Llista els dies festius i la data de validació límit del mes que se li pasa com a paràmetre
        /// </summary>
        /// <param name="year">Any</param>
        /// <param name="month">Mes</param>
        /// <returns>Llista dels dies.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<Day> listCalendar(int year, int month)
        {
            log.Debug("listCalendar Begin");
            List<Day> result = new List<Day>();
            try
            {
                DateTime min = new DateTime(year, month, 1);
                DateTime max = new DateTime(year, month, DateTime.DaysInMonth(year, month));

                //Busquem el día top de validació per el més seleccionat
                //Es fa perque per defecte el métode crea el dia a la taula si no existeix.
                getCalendarValidationLimit(year, month);

                //Seleccionem totes les entrades referents al al més seleccionat
                result.AddRange(from d in db.Calendar where d.Date >= min && d.Date <=max select d);
                log.Info("Llista de dies del més generada. " + result.Count + " elements");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar dies del més.", e);
            }
            log.Debug("listCalendar End");
            return result;
        }

        /// <summary>
        /// Recupera el dia de validacó límit per el mes que se li pasa com a paràmetre. Si no
        /// existeix una data limit per el mes es crea una per defecte.
        /// </summary>
        /// <param name="year">Any</param>
        /// <param name="month">Mes</param>
        /// <returns>Dia amb la data límit</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public Day getCalendarValidationLimit(int year, int month)
        {
            log.Debug("getCalendarValidationLimit Begin");
            Day result = new Day();
            try
            {
                DateTime min = new DateTime(year, month, 1);
                DateTime max = new DateTime(year, month, DateTime.DaysInMonth(year, month));

                //Busquem el día top de validació per el més seleccionat
                var query = (from d in db.Calendar where d.Date >= min && d.Date <= max && d.Type == DayType.VALIDATION_MAX select d);
                if (query.Count() == 0)
                {
                    setDefaultValidationLimit(year, month);
                    query = (from d in db.Calendar where d.Date >= min && d.Date <= max && d.Type == DayType.VALIDATION_MAX select d);
                }
                //Selleccionem la data límit del mes
                result = query.First();
                log.Info("Data límit " + result.Date.ToShortDateString() + " seleccionada.");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al retornar el dia máxim de validació.", e);
            }
            log.Debug("getCalendarValidationLimit End");
            return result;
        }

        /// <summary>
        /// Modifica el dia de validació límit. Si al més corresponent a la data existeix
        /// una altra data de validació límit aquesta es esborrada.
        /// </summary>
        /// <param name="dateTime">Data a marcar com dia limit</param>
        /// <returns>True si la pogut modificar, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool setCalendarValidationLimit(DateTime dateTime)
        {
            log.Debug("setCalendarValidationLimit Begin");
            bool result = false;
            try
            {
                DateTime min = new DateTime(dateTime.Year, dateTime.Month, 1);
                DateTime max = new DateTime(dateTime.Year, dateTime.Month, DateTime.DaysInMonth(dateTime.Year, dateTime.Month));

                //Busquem el día top de validació per el més seleccionat (nomès por haver-hi un)
                var query = (from d in db.Calendar where d.Date >= min && d.Date <= max && d.Type == DayType.VALIDATION_MAX select d);
                bool needToSet = false;
                if (query.Count() == 0)
                {
                    //Si no existeix el creem 
                    log.Info("No hi ha una data limit en el més que es vol afegir la data.");
                    needToSet = true;
                }
                else
                {
                    if (!query.First().Date.Equals(dateTime.Date))
                    {
                        // Si es tracta d'una altre data esborrem la anterior i afegim la nova.
                        log.Info("Ja hi ha una data limit en el més que es vol afegir la data. L'esborrem");
                        db.Calendar.Remove(query.First());
                        needToSet = true;
                    }
                }

                if (needToSet)
                {
                    Day newLimitDay = new Day { Date = dateTime.Date, Type = DayType.VALIDATION_MAX };
                    db.Calendar.Add(newLimitDay);
                    db.SaveChanges();
                }

                result = true;
                log.Info("Es seteja " + dateTime.ToShortDateString() + " com a data limit.");
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al setejar el dia màxim per validació del mes.", e);
            }
            log.Debug("setCalendarValidationLimit End");
            return result;
        }

        /// <summary>
        /// Crea un dia festiu
        /// </summary>
        /// <param name="dateTime">Data del dia festiu</param>
        /// <returns>True si ha pogut marcar el dia com festiu, False si no</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool setCalendarHoliday(DateTime dateTime)
        {
            log.Debug("setCalendarHoliday Begin");
            bool result = false;
            try
            {
                //Busquem el día festiu (nomès por haver-hi un)
                var query = (from d in db.Calendar where d.Date == dateTime.Date && d.Type == DayType.HOLIDAY select d);
                if (query.Count() == 0)
                {
                    //Si no existeix el creem 
                    log.Info("El dia seleccionat no esta marcat com a festiu...");
                    Day newHoliday = new Day() { Date = dateTime.Date, Type = DayType.HOLIDAY };
                    db.Calendar.Add(newHoliday);
                    db.SaveChanges();
                    log.Info("Es dia " + dateTime.ToShortDateString() + " ha estat marcat com a dia festiu.");
                }
                else
                {
                    // Si existeix el deixem
                    log.Info("El dia " + dateTime.ToShortDateString() + " ja era marcat com festiu.");
                }
                result = true;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al marcar un dia festiu.", e);
            }
            log.Debug("setCalendarHoliday End");
            return result;
        }

        /// <summary>
        /// Esborra un dia festiu del calendari
        /// </summary>
        /// <param name="dateTime">Data del dia festiu a esborrar</param>
        /// <returns>True si ha pogut modificar el dia, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool deleteCalendarHoliday(DateTime dateTime)
        {
            log.Debug("setCalendarHoliday Begin");
            bool result = false;
            try
            {
                //Busquem el día festiu (nomès por haver-hi un)
                var query = (from d in db.Calendar where d.Date == dateTime.Date && d.Type == DayType.HOLIDAY select d);
                if (query.Count() == 1)
                {
                    //Si existeix l'eliminem
                    db.Calendar.Remove(query.First());
                    db.SaveChanges();
                    log.Info("Es dia " + dateTime.ToShortDateString() + " ha estat eliminat com a dia festiu.");
                }
                // Intentar esborrar un dia festiu que no està marcat no ho considerem com un error
                //else
                //{ 
                //}
                result = true;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al eliminar un dia festiu.", e);
            }
            log.Debug("setCalendarHoliday End");
            return result;
        }

        /// <summary>
        /// Comproba si una data es festiva segons el calendari
        /// </summary>
        /// <param name="date">Data a comprobar si es festiu</param>
        /// <returns>Tru si el dia es festiu, False si no</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool isDateHoliday(DateTime date) 
        {
            log.Debug("isDateHoliday Begin");
            bool result = false;
            try
            {
                //Busquem el día festiu (nomès por haver-hi un)
                var query = (from d in db.Calendar where d.Date == date.Date && d.Type == DayType.HOLIDAY select d);
                if (query.Count() == 1)
                {
                    log.Info("La data " + date.Date.ToShortDateString() + " està marcat com dia festiu.");
                    result = true;
                }
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al comprobar si un dia és festiu.", e);
            }
            log.Debug("isDateHoliday End");
            return result;

        }

        /// <summary>
        /// Comproba si un dia cau en cap de setmana
        /// </summary>
        /// <param name="date">Dia a comprobar si es cap de setmana</param>
        /// <returns>True si el dia es cap de setmana, False si no</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool isDateWeekend(DateTime date)
        {
            log.Debug("isDateWeekend Begin");
            bool result = false;
            try
            {
                result = Utils.DateUtils.isDateWeekend(date);
                if (result)
                {
                    log.Info("La data " + date.Date.ToShortDateString() + " és en cap de setmana.");
                }
                else 
                {
                    log.Info("La data " + date.Date.ToShortDateString() + " no és en cap de setmana.");
                }
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al comprobar si un dia cau en cap de setmana.", e);
            }
            log.Debug("isDateWeekend End");
            return result;
        }
        
        /// <summary>
        /// Marca la data limit de validacio del més que se li pasa com a parametre segons 
        /// el contingut de la taula de propietats del sistema, i sino un valor per defecte.
        /// </summary>
        /// <param name="year">Any</param>
        /// <param name="month">Mes</param>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public void setDefaultValidationLimit(int year, int month)
        {
            DateTime max = new DateTime(year, month, DateTime.DaysInMonth(year, month));

            log.Debug("Data límit per el mes seleccionat no existeix. Es crea una nova a partir del la configuració del sistema...");
            string sthreshold = getSetting(Constants.VALIDATION_THRESHOLD_KEY);
            int threshold = (sthreshold == null) ? 0 : Convert.ToInt32(sthreshold);
            if (threshold < DateTime.DaysInMonth(year, month))
            {
                //Si el threshold és lògic el restem de l'últim dia del mes
                setCalendarValidationLimit(max.AddDays(-1 * threshold));
            }
            else
            {
                //Si threshold no es un valor lògic l'últim día del mes és la data límit
                log.Debug("Configuració del sistema ilògica. Es considera l'últim dia del més com a data limit.");
                setCalendarValidationLimit(max);
            }
        }
    }
}