﻿using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using GestorDespeses.Exceptions;
using GestorDespeses.Model;

namespace GestorDespeses.BI
{
    public class ExpensesBI
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private GestorDespesesContext db;
        private bool initializedDB;

        public ExpensesBI()
        {
            log.Info("Initializing database context...");
            db = new GestorDespesesContext();
            initializedDB = true;
            log.Info("Context initialized");
        }

        public ExpensesBI(GestorDespesesContext context)
        {
            db = context;
            initializedDB = false;
        }

        ~ExpensesBI()
        {
            if (initializedDB)
            {
                log.Info("Disposing database context...");
                db.Dispose();
                log.Info("Context disposed");
            }
        }

        /// <summary>
        /// Recupera un llistat amb els fulls de despeses corresponents al usuari amb userId i que estiguin
        /// en algun estat dels estats que se li pasen com a paràmetre.
        /// </summary>
        /// <param name="userId">Identificador d'usuari</param>
        /// <param name="states">Estats</param>
        /// <returns>Llistat de fulls de despeses.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<ExpensesSheet> listExpensesSheetOfUser(int userId, List<ExpensesSheetState> states)
        {
            log.Debug("listExpensesSheetOfUser Begin");
            List<ExpensesSheet> result = new List<ExpensesSheet>();
            try
            {
                //UserBI userManager = new UserBI(db);
                //User user = userManager.getUser(userId);

                // Seleccionem els fulls de despeses segons el seu estat.
                result.AddRange(from e in db.ExpensesSheets where e.UserUserId==userId && states.Contains(e.State) select e);
                log.Info("Llista de fulls de despeses generada. " + result.Count + " elements");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar fulls de despeses d'un usuari.", e);
            }
            log.Debug("listExpensesSheetOfUser End");
            return result;
        }

        /// <summary>
        /// Recupera un llistat amb els fulls de despeses corresponents al usuari amb userId i que estiguin
        /// en algun estat dels estats que se li pasen com a paràmetre.
        /// </summary>
        /// <param name="userId">Identificador d'usuari</param>
        /// <param name="states">Estats</param>
        /// <returns>Llistat de fulls de despeses.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<ExpensesSheet> listExpensesSheetToValidate(int headId, List<ExpensesSheetState> states)
        {
            log.Debug("listExpensesSheetToValidate Begin");
            List<ExpensesSheet> result = new List<ExpensesSheet>();
            try
            {
                UserBI userManager = new UserBI(db);
                User user = userManager.getUser(headId);
                if (user.Workers != null && user.Workers.Count > 0)
                {
                    // Seleccionem els fulls de despeses segons el seu estat.
                    result.AddRange(from e in db.ExpensesSheets
                                    join u in db.Users on e.UserUserId equals u.UserId
                                    where u.UserHeadId == headId && states.Contains(e.State) 
                                    select e);
                    //result.AddRange(from e in db.ExpensesSheets where user.Workers.Contains(e.User) && states.Contains(e.State) select e);
                    log.Info("Llista de fulls de despeses per validar generada. " + result.Count + " elements");
                }
                else
                {
                    log.Info("Llista de fulls de despeses per validar generada. " + result.Count + " elements");
                }
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar fulls de despeses per validar d'un responsable.", e);
            }
            log.Debug("listExpensesSheetToValidate End");
            return result;
        }

        /// <summary>
        /// Recupera un llistat amb els fulls de despeses que estiguin
        /// en algun estat dels estats que se li pasen com a paràmetre.
        /// </summary>
        /// <param name="states">Estats</param>
        /// <returns>Llistat de fulls de despeses.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<ExpensesSheet> listExpensesSheets(List<ExpensesSheetState> states)
        {
            log.Debug("listExpensesSheets Begin");
            List<ExpensesSheet> result = new List<ExpensesSheet>();
            try
            {
                result.AddRange(from e in db.ExpensesSheets where states.Contains(e.State) select e);
                log.Info("Llista de fulls de despeses generada. " + result.Count + " elements");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar fulls de despeses.", e);
            }
            log.Debug("listExpensesSheets End");
            return result;
        }

        /// <summary>
        /// Recupera la imatge de la factura de la despesa amb l'identificador que se li pasa com a parametre
        /// </summary>
        /// <param name="expenseHandler">Identificador de la despesa</param>
        /// <returns>Imatge RAW</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public byte[] getExpenseInvoice(int expenseHandler)
        {
            log.Debug("getExpenseInvoice Begin");
            byte[] result = null;
            try
            {
                // Seleccionem la imatge del full de despeses.
                Expense expense = getExpense(expenseHandler);
                result = expense.InvoiceImage;
                log.Info("Imatge recuperada. " + result.Count() + " bytes");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error recuperar la imatge d'una despesa.", e);
            }
            log.Debug("getExpenseInvoice End");
            return result;
        }

        /// <summary>
        /// Recupera una despesa per el seu identificador que es pasa com a paràmetre
        /// </summary>
        /// <param name="expenseId">Identificador de la despesa</param>
        /// <returns>Despesa</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public Expense getExpense(int expenseId)
        {
            log.Debug("getExpense Begin");
            Expense result = null;
            try
            {
                // Seleccionem la despesa. Per id nomes hi pot existir una.
                var request = (from e in db.Expenses where e.ExpensesId == expenseId select e);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Despesa amb id " + expenseId + " no existeix");
                }
                result = request.First();
                log.Info("Despesa amb id " + expenseId + " seleccionada");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error recuperar una despesa.", e);
            }
            log.Debug("getExpense End");
            return result;
        }

        /// <summary>
        /// Afegeix una nova despesa al full de despeses obert del usuari userId.
        /// </summary>
        /// <param name="userId">Identificador del usuari.</param>
        /// <param name="value">Valor/Quantitat de la despesa.</param>
        /// <param name="date">Data en que es va produir la despesa.</param>
        /// <param name="description">Descripció de la despesa</param>
        /// <param name="image">Imatge RAW amb el tiquet de la despesa.</param>
        /// <param name="expenseTypeId">Identificador del tipus de despesa.</param>
        /// <returns>Identificador de la nova despesa. -1 si no s'ha pogut crear.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public int addExpense(int userId, decimal value, DateTime date, string description, byte[] image, int expenseTypeId)
        {
            log.Debug("addExpense Begin");
            int result = -1;
            try
            {
                // Només es poden inserir despeses si l'usuari te un full de despeses obert.
                List<ExpensesSheet> sheets = listExpensesSheetOfUser(userId, new List<ExpensesSheetState>() { ExpensesSheetState.OPEN });
                if (sheets.Count() != 1)
                {
                    throw new GestorDespesesException("L'usuari no te cap full de despeses obert.");
                }
                ExpensesSheet expenseSheetOpen = sheets.First();
                // Recuperem el tipus de despesa.
                ExpensesTypesBI expensesTypesBIManager = new ExpensesTypesBI(db);
                ExpenseType expenseType = expensesTypesBIManager.getExpenseType(expenseTypeId);
                //Creem la nova despesa basant-nos en el tipus de despesa
                if (expenseType.Type.Equals(ExpenseTypeType.FIXED))
                    value = expenseType.Value.Value;
                log.Debug("Creant despensa nova...");
                CoreBI coreManager = new CoreBI(db);
                //Analitzem la data
                date = date.Date;
                decimal dateIncrement = 0;
                bool dateIsOnHoliday = false;
                if (coreManager.isDateHoliday(date))
                {
                    dateIsOnHoliday = true;
                    dateIncrement = Convert.ToInt32(coreManager.getSetting(Constants.HOLIDAY_INCREMENT_KEY));
                }
                else
                {
                    if (coreManager.isDateWeekend(date))
                    {
                        dateIncrement = Convert.ToInt32(coreManager.getSetting(Constants.WEEKEND_INCREMENT_KEY));
                    }
                }

                // Comprobem els valor màxim
                decimal expenseTotalValue = CalculateExpenseValue(expenseType.Type, expenseType.Value, value, dateIncrement); ;
                if (expenseType.MaxValue != null &&
                    expenseTotalValue > expenseType.MaxValue)
                {
                    throw new GestorDespesesException("S'ha superat el valor màxim (" + expenseType.MaxValue + ") permès per aquest tipus de despesa.");
                }

                //Comprobem que hi ha factura si es precisa.
                if ((image == null || image.Length == 0) &&
                    expenseType.InvoiceThreshold != null &&
                    expenseTotalValue >= expenseType.InvoiceThreshold)
                {
                    throw new GestorDespesesException("No s'ha proporcinat imatge de factura i és necessària a partir de " + expenseType.InvoiceThreshold + " euros.");
                }

                var expense = new Expense
                {
                    Type = expenseType.Type,
                    TypeConcept = expenseType.Concept,
                    TypeValue = expenseType.Value,
                    TypeMaxValue = expenseType.MaxValue,
                    TypeInvoiceThreshold = expenseType.InvoiceThreshold,
                    Value = value,
                    Description = description,
                    Date = date,
                    DateIncrement = dateIncrement,
                    IsOnHolidays = dateIsOnHoliday,
                    InvoiceImage = image
                };

                // Afegim la despesa al full de despeses obert
                expenseSheetOpen.Expense.Add(expense);

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = expenseType.ExpenseTypeId;
                log.Info("Nova despesa amb id " + result + " creada");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al crear una nova despesa.", e);
            }
            log.Debug("addExpense End");
            return result;
        }

        /// <summary>
        /// Modifica el contingut d'una despesa. Si el tipus de despesa utilitzat ha cambiat s'actualitzen les dades de la despesa amb
        /// els valors del tipus de despesa seleccionat.
        /// </summary>
        /// <param name="userId">Identificador del usuari.</param>
        /// <param name="expenseId">Identificador de la despesa a editar.</param>
        /// <param name="value">Valor/Quantitat de la despesa.</param>
        /// <param name="date">Data en que es va produir la despesa.</param>
        /// <param name="description">Descripció de la despesa</param>
        /// <param name="image">Imatge RAW amb el tiquet de la despesa.</param>
        /// <param name="expenseTypeId">Identificador del tipus de despesa.</param>
        /// <returns>True si la despesa ha pogut ser modificada. False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool editExpense(int userId, int expenseId, decimal value, DateTime date, string description, byte[] image, int expenseTypeId)
        {
            log.Debug("editExpense Begin");
            bool result = false;
            try
            {
                // Recuperem la despesa a modificar
                Expense expense = getExpense(expenseId);
                // Recuperem el tipus de despesa.
                ExpensesTypesBI expensesTypesBIManager = new ExpensesTypesBI(db);
                ExpenseType expenseType = expensesTypesBIManager.getExpenseType(expenseTypeId);
                //Creem la nova despesa basant-nos en el tipus de despesa
                log.Debug("Modificant despensa ...");
                if (expenseType.Type.Equals(ExpenseTypeType.FIXED))
                    value = expenseType.Value.Value;
                CoreBI coreManager = new CoreBI(db);
                //Analitzem la data
                date = date.Date;
                decimal dateIncrement = 0;
                bool dateIsOnHoliday = false;
                if (coreManager.isDateHoliday(date))
                {
                    dateIsOnHoliday = true;
                    dateIncrement = Convert.ToInt32(coreManager.getSetting(Constants.HOLIDAY_INCREMENT_KEY));
                }
                else
                {
                    if (coreManager.isDateWeekend(date))
                    {
                        dateIncrement = Convert.ToInt32(coreManager.getSetting(Constants.WEEKEND_INCREMENT_KEY));
                    }
                }

                // Comprobem els valor màxim
                decimal expenseTotalValue = CalculateExpenseValue(expenseType.Type, expenseType.Value, value, dateIncrement); ;
                if (expenseType.MaxValue != null &&
                    expenseTotalValue > expenseType.MaxValue)
                {
                    throw new GestorDespesesException("S'ha superat el valor màxim (" + expenseType.MaxValue + ") permès per aquest tipus de despesa.");
                }

                //Comprobem que hi ha factura si es precisa.
                if ((image == null || image.Length == 0) &&
                    (expense.InvoiceImage == null || expense.InvoiceImage.Length == 0) &&
                    expenseType.InvoiceThreshold != null &&
                    expenseTotalValue >= expenseType.InvoiceThreshold)
                {
                    throw new GestorDespesesException("No s'ha proporcinat imatge de factura i és necessària a partir de " + expenseType.InvoiceThreshold + " euros.");
                }

                //Modifiquem les dades de la despesa.
                expense.Type = expenseType.Type;
                expense.TypeConcept = expenseType.Concept;
                expense.TypeValue = expenseType.Value;
                expense.TypeMaxValue = expenseType.MaxValue;
                expense.TypeInvoiceThreshold = expenseType.InvoiceThreshold;
                expense.Value = value;
                expense.Description = description;
                expense.Date = date;
                expense.DateIncrement = dateIncrement;
                expense.IsOnHolidays = dateIsOnHoliday;
                if (image == null || (image != null && image.Length > 0))
                    expense.InvoiceImage = image;

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = true;
                log.Info("Despesa amb id " + result + " modificada");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al editar una despesa.", e);
            }
            log.Debug("editExpense End");
            return result;
        }
        
        /// <summary>
        /// Esborra la despesa corresponent a l'identificador que se li pasa com a paràmetre.
        /// </summary>
        /// <param name="expenseId">Identificador de la despesa.</param>
        /// <returns>True si s'ha pogut esborrar la despesa.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool deleteExpense(int expenseId)
        {
            log.Debug("deleteExpense Begin");
            bool result = false;
            try
            {
                Expense expense = db.Expenses.Remove(getExpense(expenseId));
                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = true;
                log.Info("Despesa [" + expense.TypeConcept + "] amb id " + expenseId + " esborrada");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al borrar la despesa " + expenseId + ".", e);
            }
            log.Debug("deleteExpense End");
            return result;
        }

        /// <summary>
        /// Recupera un full de despeses de la base de dades.
        /// </summary>
        /// <param name="expensesSheetId">Identificador del full de despeses</param>
        /// <returns>Full de despeses</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public ExpensesSheet getExpensesSheet(int expensesSheetId)
        {
            log.Debug("getExpensesSheet Begin");
            ExpensesSheet result = null;
            try
            {
                // Seleccionem la despesa. Per id nomes hi pot existir una.
                var request = (from e in db.ExpensesSheets where e.ExpensesSheetId == expensesSheetId select e);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Full de despeses amb id " + expensesSheetId + " no existeix");
                }
                result = request.First();
                log.Info("Full de despeses amb id " + expensesSheetId + " seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al recuperar un full de despeses.", e);
            }
            log.Debug("getExpensesSheet End");
            return result;
        }

        /// <summary>
        /// Canvia l'estat d'un full de despeses.
        /// </summary>
        /// <param name="expensesSheetId">Identificador del full de despeses.</param>
        /// <param name="state">Nou estat del full de despeses (OPEN/VALIDATING/VALIDATED/PAYED/REVOKED)</param>
        /// <returns>True si l'estat ha sigut modificat.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool changeExpensesSheetState(int expensesSheetId, ExpensesSheetState state, string revokedReason)
        {
            log.Debug("changeExpensesSheetState Begin");
            bool result = false;
            try
            {
                // Seleccionem la despesa. Per id nomes hi pot existir una.
                ExpensesSheet expensesSheet = getExpensesSheet(expensesSheetId);
                if (expensesSheet.State == state)
                {
                    log.Info("Full de despeses amb id " + expensesSheetId + " ja estaba en estat " + state);
                }
                else
                {
                    // Control de la màquina d'estats
                    if (expensesSheet.State == ExpensesSheetState.OPEN)
                    {
                        if (state != ExpensesSheetState.VALIDATING)
                        {
                            throw new GestorDespesesException("No es pot pasar d'estat " + expensesSheet.State + " a estat " + state.ToString());
                        }
                    }
                    else if (expensesSheet.State == ExpensesSheetState.VALIDATING)
                    {
                        if (state != ExpensesSheetState.REVOKED && state != ExpensesSheetState.VALIDATED)
                        {
                            throw new GestorDespesesException("No es pot pasar d'estat " + expensesSheet.State + " a estat " + state.ToString());
                        }
                        if (String.IsNullOrEmpty(revokedReason) && state.Equals(ExpensesSheetState.REVOKED))
                        {
                            throw new GestorDespesesException("Per poder rebutjar un full de despeses s'ha de indicar una rao.");
                        }
                    }
                    else if (expensesSheet.State == ExpensesSheetState.REVOKED)
                    {
                        if (state != ExpensesSheetState.OPEN)
                        {
                            throw new GestorDespesesException("No es pot pasar d'estat " + expensesSheet.State + " a estat " + state.ToString());
                        }
                    }
                    else if (expensesSheet.State == ExpensesSheetState.VALIDATED)
                    {
                        if (state != ExpensesSheetState.PAYED)
                        {
                            throw new GestorDespesesException("No es pot pasar d'estat " + expensesSheet.State + " a estat " + state.ToString());
                        }
                    }
                    else if (expensesSheet.State == ExpensesSheetState.PAYED)
                    {
                        throw new GestorDespesesException("No es pot modificar l'estat d'un full de despeses pagat.");
                    }
                    // Si arriba aqui és que el cambi es correcte, per tant el fem.
                    expensesSheet.LastStateChangedDate = DateTime.Now;
                    expensesSheet.State = state;
                    if (state.Equals(ExpensesSheetState.REVOKED))
                    {
                        expensesSheet.RevokedReason = revokedReason;
                    }
                    if (state.Equals(ExpensesSheetState.VALIDATED))
                    {
                        // Si pasem a estat validat hem de crear un nou full de despeses per l'usuari.
                        //Creem un full de despeses en blanc i l'associem al user del propi full de despeses.
                        log.Debug("Creant un full de despeses nou per l'usuari nou...");
                        var expensesSheetOpen = new ExpensesSheet
                        {
                            LastStateChangedDate = DateTime.Now,
                            State = ExpensesSheetState.OPEN
                        };
                        expensesSheet.User.ExpensesSheets.Add(expensesSheetOpen);
                    }
                    log.Debug("Persistint els canvis...");
                    db.SaveChanges();
                    log.Debug("Canvis guardats.");

                    log.Info("Full de despeses amb id " + expensesSheetId + " cambiat a estat " + state);
                }
                result = true;
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al recuperar un full de despeses.", e);
            }
            log.Debug("changeExpensesSheetState End");
            return result;
        }

        /// <summary>
        /// Recupera un llistat amb les despeses que hi ha dins d'un full de despeses.
        /// </summary>
        /// <param name="expensesSheetId">Identificador del full de despeses.</param>
        /// <returns>Llistat de despeses</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<Expense> listExpensesOfExpensesSheet(int expensesSheetId)
        {
            log.Debug("listExpensesOfExpensesSheet Begin");
            List<Expense> result = new List<Expense>();
            try
            {
                //Seleccionem tots el usuaris
                ExpensesSheet expensesSheet = getExpensesSheet(expensesSheetId);
                var query = (from e in expensesSheet.Expense select e);
                result.AddRange(query);
                log.Info("Llista de despeses d'un full de despeses generada. " + result.Count + " elements");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar les despeses d'un full de despeses.", e);
            }
            log.Debug("listExpensesOfExpensesSheet End");
            return result;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="expenseId"></param>
        /// <param name="expenseTypeId"></param>
        /// <returns></returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool checkExpenseAgainstExpenseType(int expenseId, int expenseTypeId)
        {
            throw new NotImplementedException();
        }

        #region static members

        public static decimal ExpenseTotalValue(Expense expense)
        {
            return CalculateExpenseValue(expense.Type, expense.TypeValue, expense.Value, expense.DateIncrement);
        }

        public static decimal CalculateExpenseValue(ExpenseTypeType expenseType, decimal? typeValue, decimal value, decimal dateIncrement) {
            decimal result = 0;
            switch (expenseType)
            {
                case ExpenseTypeType.FIXED:
                    result = (decimal)typeValue;
                    break;
                case ExpenseTypeType.MULTIPLICITY:
                    result = (decimal)typeValue * value;
                    break;
                case ExpenseTypeType.VARIABLE:
                    result = value;
                    break;
            }
            result = result + ((result * dateIncrement)/100);
            return result;
        }


        #endregion
    }
}