﻿using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using GestorDespeses.Exceptions;
using GestorDespeses.Model;

namespace GestorDespeses.BI
{
    public class ExpensesTypesBI
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private GestorDespesesContext db;
        private bool initializedDB;

        public ExpensesTypesBI()
        {
            log.Debug("Initializing database context...");
            db = new GestorDespesesContext();
            initializedDB = true;
            log.Debug("Context initialized");
        }

        public ExpensesTypesBI(GestorDespesesContext context)
        {
            db = context;
            initializedDB = false;
        }

        ~ExpensesTypesBI()
        {
            if (initializedDB)
            {
                log.Debug("Disposing database context...");
                db.Dispose();
                log.Info("Context disposed");
            }
        }

        /// <summary>
        /// Recupera la llista de tots els tipus de despesa existents a la plataforma
        /// </summary>
        /// <returns>Llistat de tipus de despesa</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<ExpenseType> listExpenseTypes()
        {
            log.Debug("listExpenseTypes Begin");
            List<ExpenseType> result = new List<ExpenseType>();
            try
            {
                //Seleccionem tots el tipus de despeses
                result.AddRange(from u in db.ExpenseTypes select u);
                log.Info("Llista de tipus de despeses generada. " + result.Count + " elements");
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar tipus de despeses.", e);
            }
            log.Debug("listExpenseTypes End");
            return result;
        }

        /// <summary>
        /// Recupera la llista d'aquells tipus de despesa que el seu concepte contingui els caracters de q
        /// </summary>
        /// <param name="q">String de cerca</param>
        /// <param name="limit">Número màxim de resultats. 0 no hi ha limit.</param>
        /// <returns>Llistat de tipus de despesa</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<ExpenseType> listExpenseTypesByQuery(string q, int limit)
        {
            log.Debug("listExpenseTypesByQuery Begin");
            List<ExpenseType> result = new List<ExpenseType>();
            try
            {
                q = q.ToLower();
                //Seleccionem els tipus de despeses que en el concepte continguin la cadena q.
                var query = (from u in db.ExpenseTypes where u.Concept.ToLower().Contains(q) select u);
                if (limit > 0)
                {
                    result.AddRange(query.Take(limit));
                }
                else
                {
                    result.AddRange(query);
                }
                log.Info("Llista de tipus de despeses generada. " + result.Count + " elements");
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar tipus de despeses amb paràmetres.", e);
            }
            log.Debug("listExpenseTypesByQuery End");
            return result;
        }

        /// <summary>
        /// Recupera un tipus de despesa segons el seu identificador passat com a paràmetre
        /// </summary>
        /// <param name="espenseTypeId">Identificador de tipus de despesa</param>
        /// <returns>Tipus de despesa</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public ExpenseType getExpenseType(int espenseTypeId)
        {
            log.Debug("getExpenseType Begin");
            ExpenseType result = null;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from u in db.ExpenseTypes
                               where u.ExpenseTypeId == espenseTypeId
                               select u);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Tipus de despesa amb id " + espenseTypeId + " no existeix");
                }
                result = request.First();
                log.Info("Tipus de despesa id " + result.ExpenseTypeId + " seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un tipus de despesa.", e);
            }
            log.Debug("getExpenseType End");
            return result;
        }

        /// <summary>
        /// Recupera un tipus de despesa segons el seu identificador passat com a paràmetre
        /// </summary>
        /// <param name="espenseTypeId">Identificador de tipus de despesa</param>
        /// <returns>Tipus de despesa</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public ExpenseType getExpenseTypeByConcept(string espenseTypeConcept)
        {
            log.Debug("getExpenseTypeByConcept Begin");
            ExpenseType result = null;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from u in db.ExpenseTypes
                               where u.Concept == espenseTypeConcept
                               select u);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Tipus de despesa amb concepte " + espenseTypeConcept + " no existeix");
                }
                result = request.First();
                log.Info("Tipus de despesa cancepte " + result.Concept + " seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un tipus de despesa.", e);
            }
            log.Debug("getExpenseTypeByConcept End");
            return result;
        }

        /// <summary>
        /// Afageix un nou tipus de despesa al sistema amb les dades que se li pasen com a paràmetre
        /// </summary>
        /// <param name="concept">Cencepte de la despesa</param>
        /// <param name="value">Valor/Valor per unitat</param>
        /// <param name="maxValue">Valor màxim de la despesa</param>
        /// <param name="threshold">Valor a partir del qual s'ha d'incloure factura</param>
        /// <param name="type">Tipus de despesa(FIXED/MULTIPLICITY/VARIABLE)</param>
        /// <returns>L'identificador intern del nou tipus de despesa o -1 si no s'ha pogut crear</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public int addExpenseType(string concept, decimal? value, decimal? maxValue, decimal? threshold, ExpenseTypeType type)
        {
            log.Debug("addExpenseType Begin");
            int result = -1;
            try
            {
                CheckExpenseType(ref value, ref maxValue, ref threshold, type);

                //Creem el nou tipus de despesa
                log.Debug("Creant tipus de despensa nou...");
                var expenseType = new ExpenseType
                {
                    Concept = concept,
                    Value = value,
                    MaxValue = maxValue,
                    InvoiceThreshold = threshold,
                    Type = type
                };
                expenseType = db.ExpenseTypes.Add(expenseType);

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = expenseType.ExpenseTypeId;
                log.Info("Nou tipus de despesa amb id " + result + " creat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al crear un nou tipus de despesa.", e);
            }
            log.Debug("addExpenseType End");
            return result;
        }

        /// <summary>
        /// Esborra un tipus de despesa especificat per el identificador que es pasa com a paràmetre
        /// </summary>
        /// <param name="expenseTypeId">Identicador del tipus de despesa a esborrar</param>
        /// <returns>True si s'ha pogut esborrat correctament. False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool deleteExpenseType(int expenseTypeId)
        {
            log.Debug("deleteExpenseType Begin");
            bool result = false;
            try
            {
                ExpenseType expenseType = db.ExpenseTypes.Remove(getExpenseType(expenseTypeId));
                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = true;
                log.Info("Tipus de despesa [" + expenseType.Concept + "] amb id " + expenseTypeId + " esborrat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al borrar el tipus de despesa " + expenseTypeId + ".", e);
            }
            log.Debug("deleteExpenseType End");
            return result;
        }

        /// <summary>
        /// Modifica els valors de la despesa especificada per l'identificador que es pasa com a paràmetre.
        /// </summary>
        /// <param name="expenseTypeId">Identicador del tipus de despesa a editar</param>
        /// <param name="concept">Cencepte de la despesa</param>
        /// <param name="value">Valor/Valor per unitat</param>
        /// <param name="maxValue">Valor màxim de la despesa</param>
        /// <param name="threshold">Valor a partir del qual s'ha d'incloure factura</param>
        /// <param name="type">Tipus de despesa(FIXED/MULTIPLICITY/VARIABLE)</param>
        /// <returns>True si s'ha pogut editar. False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool editExpenseType(int expenseTypeId, string concept, decimal? value, decimal? maxValue, decimal? threshold, ExpenseTypeType type)
        {
            log.Debug("editExpenseType Begin");
            bool result = false;
            try
            {
                CheckExpenseType(ref value, ref maxValue, ref threshold, type);

                ExpenseType expenseType = getExpenseType(expenseTypeId);
                expenseType.Concept = concept;
                expenseType.Value = value;
                expenseType.MaxValue = maxValue;
                expenseType.InvoiceThreshold = threshold;
                expenseType.Type = type;

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");
                result = true;
                log.Info("Tipus de despesa amb id " + expenseTypeId + " modifcat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al editar un tipus de despesa.", e);
            }
            log.Debug("editExpenseType End");
            return result;
        }

        #region private

        /// <summary>
        /// Comprova que els valors de la despesa són valids per el seu tipus.
        /// </summary>
        /// <param name="value">Valor/Valor per unitat</param>
        /// <param name="maxValue">Valor màxim de la despesa</param>
        /// <param name="threshold">Valor a partir del qual s'ha d'incloure factura</param>
        /// <param name="type">Tipus de despesa(FIXED/MULTIPLICITY/VARIABLE)</param>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        private static void CheckExpenseType(ref decimal? value, ref decimal? maxValue, ref decimal? threshold, ExpenseTypeType type)
        {
            log.Debug("Comprobant tipus de despesa...");

            //Fem comprobacio dels paràmetres d'entradasiguin lògics
            switch (type)
            {
                case ExpenseTypeType.FIXED:
                    //Si es tracta d'un valor fix MaxValue i Threshold han de ser -1
                    log.Debug("Despesa de tipus FIXED");
                    if (value != null && value > 0)
                    {
                        maxValue = threshold = null;
                    }
                    else
                    {
                        throw new GestorDespesesException("Quan type=FIXED el valor de value ha de ser superior a 0.");
                    }
                    break;
                case ExpenseTypeType.MULTIPLICITY:
                    log.Debug("Despesa de tipus MULTIPLICITY");
                    //Si es tracta d'un valor Multipicatiu (€/km) s'han de comprobar el valors de value, maxValue i threshold
                    if (value != null && value > 0)
                    {
                        if (maxValue != null && maxValue <= 0)
                        {
                            throw new GestorDespesesException("Quan type=MULTIPLICITY el valor de maxValue ha de ser null o ha de ser superior a 0.");
                        }
                        if (threshold != null && threshold <= 0)
                        {
                            throw new GestorDespesesException("Quan type=MULTIPLICITY el valor de threshold ha de ser null o ha de ser superior a 0.");
                        }
                        if (maxValue != null && threshold != null && maxValue < threshold)
                        {
                            throw new GestorDespesesException("Quan type=MULTIPLICITY els valors de maxValue i threshold han de ser nulls o han de ser superiors a 0, i threshold mai pot ser superior a maxValue.");
                        }
                    }
                    else
                    {
                        throw new GestorDespesesException("Quan type=MULTIPLICITY el valor de value ha de ser superior a 0.");
                    }
                    break;
                case ExpenseTypeType.VARIABLE:
                    log.Debug("Despesa de tipus VARIABLE");
                    //Si es tracta d'un valor variable s'han de comprobar el valors de maxValue i threshold
                    if (maxValue != null && maxValue <= 0)
                    {
                        throw new GestorDespesesException("Quan type=VARIABLE el valor de maxValue ha de ser null o ha de ser superior a 0.");
                    }
                    if (threshold != null && threshold <= 0)
                    {
                        throw new GestorDespesesException("Quan type=VARIABLE el valor de threshold ha de ser null o ha de ser superior a 0.");
                    }
                    if (maxValue != null && threshold != null && maxValue < threshold)
                    {
                        throw new GestorDespesesException("Quan type=VARIABLE els valors de maxValue i threshold han de ser nulls o han de ser superiors a 0, i threshold mai pot ser superior a maxValue.");
                    }
                    value = null;
                    break;
            }
        }
        #endregion
    }
}