﻿using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Transactions;
using System.Web;
using GestorDespeses.Exceptions;
using GestorDespeses.Utils;
using GestorDespeses.Model;
using System.Data;
using System.Data.SqlClient;
using System.Data.Entity.Infrastructure;

namespace GestorDespeses.BI
{
    public class UserBI
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private GestorDespesesContext db;
        private bool initializedDB;

        public UserBI()
        {
            log.Debug("Initializing database context...");
            db = new GestorDespesesContext();
            initializedDB = true;
            log.Debug("Context initialized");
        }

        public UserBI(GestorDespesesContext context)
        {
            db = context;
            initializedDB = false;
        }

        ~UserBI()
        {
            if (initializedDB)
            {
                log.Debug("Disposing database context...");
                db.Dispose();
                log.Info("Context disposed");
            }
        }

        /// <summary>
        /// Recupera la llista de tots els usuaris donats d'alta a la plataforma
        /// </summary>
        /// <returns>LListat d'usuaris</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<User> listUsers()
        {
            log.Debug("listUsers Begin");
            List<User> result = new List<User>();
            try
            {
                //Seleccionem tots el usuaris
                result.AddRange(from u in db.Users select u);
                log.Info("Llista d'usuaris generada. " + result.Count + " elements");
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar usuaris.", e);
            }
            log.Debug("listUsers End");
            return result;
        }

        /// <summary>
        /// Recupera la llista d'aquells usuaris que el seu username, full name o email contenen els caracters de q
        /// </summary>
        /// <param name="q">String de cerca</param>
        /// <param name="limit">Número màxim de resultats. 0 no hi ha limit.</param>
        /// <returns>Llistat d'usuaris</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public List<User> listUsersByQuery(string q, int limit)
        {
            log.Debug("listUsersByQuery Begin");
            List<User> result = new List<User>();
            try
            {
                q = q.ToLower();
                //Seleccionem els usuaris que en el username, full name o email continguin la cadena q.
                var query = (from u in db.Users where u.Username.ToLower().Contains(q) || u.FullName.ToLower().Contains(q) || u.Email.ToLower().Contains(q) select u);
                if (limit > 0)
                {
                    result.AddRange(query.Take(limit));
                }
                else 
                {
                    result.AddRange(query);
                }
                log.Info("Llista d'usuaris generada. " + result.Count + " elements");
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al llistar usuaris amb paràmetres.", e);
            }
            log.Debug("listUsersByQuery End");
            return result;
        }

        /// <summary>
        /// Recupera l'usuari amb el identificador que se li pasa com a paràmetre
        /// </summary>
        /// <param name="userId">Identificador</param>
        /// <returns>L'usuari si el troba, null si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public User getUser(int userId)
        {
            log.Debug("getUser Begin");
            User result = null;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from u in db.Users
                        where u.UserId == userId
                        select u);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Usuari amb id " + userId + " no existeix");
                }
                result = request.First();
                log.Info("Usuari amb id " + result.UserId + " seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un usuari.", e);
            }
            log.Debug("getUser End");
            return result;
        }

        /// <summary>
        /// Recupera l'usuari amb el username que se li pasa com a paràmetre
        /// </summary>
        /// <param name="username">Username</param>
        /// <returns>L'usuari si el troba, null si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public User getUserByUsername(string username)
        {
            log.Debug("getUserByUsername Begin");
            User result = null;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from u in db.Users
                          where u.Username == username
                          select u);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Usuari amb username " + username + " no existeix.");
                }
                result = request.First();
                log.Info("Usuari amb username " + result.Username + " seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un usuari per username.", e);
            }
            log.Debug("getUserByUsername End");
            return result;
        }

        /// <summary>
        /// Recupera l'usuari amb la direcció de correu electronic que se li pasa com a paràmetre
        /// </summary>
        /// <param name="email">Email</param>
        /// <returns>L'usuari si el troba, null si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public User getUserByEmail(string email)
        {
            log.Debug("getUserByEmail Begin");
            User result = null;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from u in db.Users
                          where u.Email == email
                          select u);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Usuari amb email " + email + " no existeix.");
                }
                result = request.First();
                log.Info("Usuari amb email " + result.Email + " seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un usuari per email.", e);
            }
            log.Debug("getUserByEmail End");
            return result;
        }

        /// <summary>
        /// Recupera l'usuari que té el codi d'activació que se li pasa com a paràmetre i que aquest
        /// codi no fa més de dues hores que es va generar.
        /// </summary>
        /// <param name="activationCode">Codi d'activació</param>
        /// <returns>L'usuari si el troba, null si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public User getUserByActivationCode(string activationCode)
        {
            log.Debug("getUserByActivationCode Begin");
            User result = null;
            try
            {
                if (String.IsNullOrEmpty(activationCode))
                {
                    throw new GestorDespesesException("El codi d'activació no pot ser null ni estar buit.");
                }
                //Seleccionem el primer resultat (es únic).
                DateTime twoHoursAgo = DateTime.Now.Subtract(new TimeSpan(2, 0, 0));
                var request = (from u in db.Users
                                where u.ActivationCode == activationCode && u.ActivationCodeDate > twoHoursAgo
                                select u);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Usuari amb activationCode " + activationCode + " no existeix o el codi ha caducat.");
                }
                result = request.First();
                log.Info("Usuari amb activationCode " + result.ActivationCode + " actiu seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un usuari per activation code.", e);
            }
            log.Debug("getUserByActivationCode End");
            return result;
        }

        /// <summary>
        /// Afageix un usuari nou a la base de dades amb les dades que se li pasen com a paràmetre
        /// </summary>
        /// <param name="username">Usuari</param>
        /// <param name="password">Contrasenya</param>
        /// <param name="email">Email</param>
        /// <param name="fullName">Nom complet</param>
        /// <param name="type">Tipus d'usuari (WORKER/ADMINISTRATOR)</param>
        /// <returns>L'identificador del nou usuari o -1 si no s'ha pogut crear.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public int addUser(string username,
            string password,
            string email,
            string fullName,
            UserType type)
        {
            log.Debug("addUser Begin");
            int result = -1;
            try
            {
                if (!SMTPUtils.IsValidEmail(email))
                {
                    throw new GestorDespesesException("Direcció de correu no vàlida.");
                }
                if (string.IsNullOrEmpty(username)) 
                {
                    throw new GestorDespesesException("El username no pot estar vuit.");
                }
                if (string.IsNullOrEmpty(fullName))
                {
                    throw new GestorDespesesException("El nom del usuari no pot estar vuit.");
                }
                if (password == null || password.Length < 5 )
                {
                    throw new GestorDespesesException("La contrasenya ha de tenir una longitud mínima de 5 caràcters.");
                }

                //Creem un full de despeses en blanc i l'associem al user creat
                log.Debug("Creant un full de despeses nou per l'usuari nou...");
                var expensesSheet = new ExpensesSheet
                {
                    LastStateChangedDate = DateTime.Now,
                    State = ExpensesSheetState.OPEN
                };

                //Creem un usuari de despeses en blanc
                log.Debug("Creant l'usuari nou...");
                var user = new User
                {
                    Username = username,
                    Password = Utils.SHA1Util.SHA1HashStringForUTF8String(password),
                    Email = email,
                    FullName = fullName,
                    Type = type
                };

                //Assignem el full de despeses al usuari
                log.Debug("Assignant el full de despeses nou a l'usuari nou...");
                user.ExpensesSheets.Add(expensesSheet);
                user = db.Users.Add(user);

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = user.UserId;
                log.Info("Nou usuari amb id " + result + " creat");
            }
            catch (DbUpdateException ue)
            {
                SqlException innerException = ue.InnerException.InnerException as SqlException;
                if (innerException != null && innerException.Number == 2627)   
                {
                    if(innerException.Message.Contains("UC_USERS_USERNAME"))
                    {
                        throw new GestorDespesesException("El camp username(" + username + ") ja està sent usat per un altre usuari.");
                    }
                    else //if (innerException.Message.Contains("UC_USERS_EMAIL"))
                    {
                        throw new GestorDespesesException("El camp email(" + email + ") ja està sent usat per un altre usuari.");
                    }
                }
                throw new GestorDespesesUnknownException("Error a la base de dades al crear un nou usuari.", ue);   
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al crear un nou usuari.", e);
            }
            log.Debug("addUser End");
            return result;
        }

        /// <summary>
        /// Esborra l'usuari amb identificador que se li pasa com a paràmetre
        /// </summary>
        /// <param name="userId">Identificador</param>
        /// <returns>True si s'ha pogut esborrar, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool deleteUser(int userId)
        {
            log.Debug("deleteUser Begin");
            bool result = false;
            try
            {
                User user = getUser(userId);
                user.Workers.Clear();
                user = db.Users.Remove(user);
                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = true;
                log.Info("Usuari [" + user.Username + "] amb id " + userId + " esborrat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al esborrar l'usuari " + userId + ".", e);
            }
            log.Debug("deleteUser End");
            return result;
        }

        /// <summary>
        /// Modifica les dades de l'usuari amb identificador que se li pasa com a paràmetre
        /// </summary>
        /// <param name="userId">Identificador de l'usuari</param>
        /// <param name="password">Contrasenya</param>
        /// <param name="email">Email</param>
        /// <param name="fullName">Nom complet</param>
        /// <param name="type">Tipus d'usuari (WORKER/ADMINISTRATOR)</param>
        /// <returns>True si s'ha pogut modificar, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool editUser(int userId,
            string password,
            string email,
            string fullName,
            UserType type)
        {
            log.Debug("editUser Begin");
            bool result = false;
            try
            {
                if (!SMTPUtils.IsValidEmail(email))
                {
                    throw new GestorDespesesException("Direcció de correu no vàlida.");
                }
                if (string.IsNullOrEmpty(fullName))
                {
                    throw new GestorDespesesException("El nom del usuari no pot estar vuit.");
                }
                if (password != null && password.Length > 1 && password.Length < 5)
                {
                    throw new GestorDespesesException("La contrasenya ha de tenir una longitud mínima de 5 caràcters.");
                }
                User user = getUser(userId);
                //Si el password ve vuit no el modifiquem.
                if (!String.IsNullOrEmpty(password))
                {
                    user.Password = Utils.SHA1Util.SHA1HashStringForUTF8String(password);
                }
                user.Email = email;
                user.FullName = fullName;
                user.Type = type;

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");
                result = true;
                log.Info("Usuari amb id " + userId + " modifcat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un usuari.", e);
            }
            log.Debug("editUser End");
            return result;
        }

        /// <summary>
        /// Valida les credencials del usuari que se li pasa com a paràmetre
        /// </summary>
        /// <param name="username">Usuari</param>
        /// <param name="password">Contrasenya</param>
        /// <returns>True si la validació ha sigut correcte, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool validateUser(string username, string password)
        {
            log.Debug("validateUser Begin");
            bool result = false;
            try
            {
                //Seleccionem els ids del usuaris que coincideixen amb el usermane i password (nomès pot existir un per username)
                //proporcionats i contem quants n'hi ha, si hi ha 1 és que és valid.
                string passwordHash = Utils.SHA1Util.SHA1HashStringForUTF8String(password);
                result = (from u in db.Users
                          where u.Username == username && u.Password == passwordHash
                          select u.UserId).Count() == 1;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al validar un usuari.", e);
            }
            log.Debug("validateUser End");
            
            return result;
        }

        /// <summary>
        /// Assigna un usuari responsable a un usuari
        /// </summary>
        /// <param name="userId">Usuari</param>
        /// <param name="headId">Usuari responsable</param>
        /// <returns>True si s'ha pogut assignar, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool assignHead(int userId, int headId)
        {
            log.Debug("assignHead Begin");
            bool result = false;
            try
            {
                User user = getUser(userId);
                User head = getUser(headId);
                user.Head = head;
                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");
                result = true;
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al asignar un cap a un usuari.", e);
            }
            log.Debug("assignHead End");

            return result;
        }

        /// <summary>
        /// Genera un codi d'activació per l'usuari amb l'identificador que es pasa com a paràmetre i 
        /// envia un correu electronic al usuari amb el codi generat
        /// </summary>
        /// <param name="userId">Identificador</param>
        /// <returns>True si s'ha pogut generar i enviar el codi, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool generateValidationCode(int userId)
        {
            log.Debug("generateValidationCode Begin");
            bool result = false;
            try
            {
                User user = getUser(userId);
                result = generateCodeAndSendEmail(user);
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al generar un codi per l'usuari.", e);
            }
            log.Debug("generateValidationCode End");

            return result;
        }

        /// <summary>
        /// Genera un codi d'activació per l'usuari amb username o email que es pasa com a paràmetre i 
        /// envia un correu electronic al usuari amb el codi generat
        /// </summary>
        /// <param name="userId">Identificador</param>
        /// <returns>True si s'ha pogut generar i enviar el codi, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool generateValidationCode(string usernameOrEmail)
        {
            log.Debug("generateValidationCode Begin");
            bool result = false;
            try
            {
                User user = null;
                if (SMTPUtils.IsValidEmail(usernameOrEmail))
                {
                    //La entrada es tracta del email del usuari
                    user = getUserByEmail(usernameOrEmail);
                }
                else
                {
                    //La entrada es tracta del username del usuari
                    user = getUserByUsername(usernameOrEmail);
                }
                result = generateCodeAndSendEmail(user);
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al generar un codi per l'usuari.", e);
            }
            log.Debug("generateValidationCode End");

            return result;
        }

        /// <summary>
        /// Genera una nova contrasenya de manera aleatoria per l'usuari que es pasa com a paràmetre i
        /// envia un correu a l'usuari amb la nova contrasenya. Si l'enviament no funciona la contrasenya es
        /// retornada igualment per el métode.
        /// </summary>
        /// <param name="userId">Identificador</param>
        /// <returns>String que conté la nova contrasenya o null si no s'ha pogut generar.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public string resetPassword(int userId)
        {
            log.Debug("resetPassword Begin");
            CoreBI coreManager = new CoreBI(db);

            string result = null;
            try
            {
                User user = getUser(userId);
                string newPassword = Utils.RandomUtils.RandomString(12);
                user.Password = Utils.SHA1Util.SHA1HashStringForUTF8String(newPassword);
                user.ActivationCode = null;

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = newPassword;
                log.Debug("Enviant informació del password a l'usuari...");
                // Enviem un correu al usuari amb el seu nou password.
                coreManager.sendSMTP(user, Constants.PASSWORD_RESET_SUBJECT, String.Format(Constants.PASSWORD_RESET_MESSAGE, newPassword));
                log.Info("Canvi de password per l'usuari " + user.Username + " satisfactori.");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al resetejar el password.", e);
            }
            log.Debug("resetPassword End");
            return result;
        }

        /// <summary>
        /// Modifica la contrasenya de l'usuari que tingui el codi d'activació que se li pasa com a paràmetre.
        /// </summary>
        /// <param name="newPassword">Contrasenya</param>
        /// <param name="activationCode">Codi d'activació</param>
        /// <returns>True si s'ha pogut modificar la contrasenya, False si no.</returns>
        public bool changePasswordByValidationCode(string newPassword, string activationCode)
        {
            log.Debug("changePasswordByValidationCode Begin");
            bool result = false;
            try
            {
                if (newPassword == null || newPassword.Length < 5)
                {
                    throw new GestorDespesesException("La contrasenya ha de tenir una longitud mínima de 5 caràcters.");
                }
                User user = getUserByActivationCode(activationCode);
                user.Password = Utils.SHA1Util.SHA1HashStringForUTF8String(newPassword);
                user.ActivationCode = null;

                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");

                result = true;
                log.Info("Canvi de password per l'usuari " + user.Username + " satisfactori.");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                log.Error("Error al modificar el password mitjançant activationCode.\r\n" + e.StackTrace);
                throw new GestorDespesesUnknownException(".", e);
            }
            log.Debug("changePasswordByValidationCode End");
            return result;
        }

        #region private

        private bool generateCodeAndSendEmail(User user)
        {
            log.Debug("generateCodeAndSendEmail Begin");
            CoreBI coreManager = new CoreBI(db);

            bool result = false;
            if (user != null)
            {
                log.Debug("Generant nou codi...");
                // Creem el codi nou.
                user.ActivationCode = Utils.RandomUtils.RandomString(12);
                user.ActivationCodeDate = DateTime.Now;
                log.Debug("Persistint els canvis...");
                db.SaveChanges();
                log.Debug("Canvis guardats.");
                log.Info("Codi generat per l'usuari " + user.Username + ". Codi: " + user.ActivationCode);

                log.Debug("Enviant informació del codi a l'usuari...");
                // Enviem un correu al usuari amb el seu nou codi d'activació.
                coreManager.sendSMTP(user, Constants.ACTIVATION_CODE_SUBJECT, String.Format(Constants.ACTIVATION_CODE_MESSAGE, user.ActivationCode));
                result = true;
            }
            log.Debug("generateCodeAndSendEmail End");
            return result;
        }

        #endregion
    }
}