﻿using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using GestorDespeses.Exceptions;
using GestorDespeses.Model;

namespace GestorDespeses.BI
{
    public class WSSessionBI
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private GestorDespesesContext db;
        private bool initializedDB;

        public WSSessionBI()
        {
            log.Info("Initializing database context...");
            db = new GestorDespesesContext();
            initializedDB = true;
            log.Info("Context initialized");
        }

        public WSSessionBI(GestorDespesesContext context)
        {
            db = context;
            initializedDB = false;
        }

        ~WSSessionBI()
        {
            if (initializedDB)
            {
                log.Info("Disposing database context...");
                db.Dispose();
                log.Info("Context disposed");
            }
        }

        /// <summary>
        /// Crea una entrada de sessió nova a la taula de sessions
        /// </summary>
        /// <param name="userId">Identificador de l'usuari que se li crea la sessió.</param>
        /// <returns>Token de sessió generat</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public string createTokenForUser(int userId)
        {
            log.Debug("createTokenForUser Begin");
            string result = null;
            try
            {
                UserBI usersManager = new UserBI(db);
                User user = usersManager.getUser(userId);
                WSSessionToken token = new WSSessionToken()
                {
                    InitDateTime = DateTime.Now,
                    lastAccessDateTime = DateTime.Now,
                    Token = Utils.RandomUtils.RandomString(20)
                };
                // Si l'usuari ja te un token asignat l'esborrem
                if (user.WSSessionToken != null)
                {
                    db.WSSessionTokens.Remove(user.WSSessionToken);
                }
                user.WSSessionToken = token;
                db.SaveChanges();
                result = token.Token;
                log.Info("Token " + result + " per l'usuari amb id " + userId + " creat.");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al crear un token per un usuari.", e);
            }
            log.Debug("createTokenForUser End");
            return result;
        }

        /// <summary>
        /// Recupera l'usuari relacionat amb el token de sessió
        /// </summary>
        /// <param name="token">Token</param>
        /// <returns>Usuari</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public User getUserByToken(string token)
        {
            log.Debug("getUserByToken Begin");
            updateTokenLastAccessDate(token);
            User result = null;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from u in db.Users
                               where u.WSSessionToken.Token == token
                               select u);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Usuari amb token " + token + " no existeix.");
                }
                result = request.First();
                log.Info("Usuari amb token " + token + " seleccionat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al cercar un usuari per token.", e);
            }
            log.Debug("getUserByToken End");
            return result;
        }

        /// <summary>
        /// Esborra la sessió del usuari amb identificador userId
        /// </summary>
        /// <param name="userId">Identificador d'usuari</param>
        /// <returns>True si ha pogut esborrar la sessio, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool deleteUserToken(int userId)
        {
            log.Debug("deleteUserToken Begin");
            bool result = false;
            try
            {
                UserBI usersManager = new UserBI(db);
                User user = usersManager.getUser(userId);
                if (user.WSSessionToken != null)
                {
                    WSSessionToken token = user.WSSessionToken;
                    db.WSSessionTokens.Remove(token);
                    db.SaveChanges();
                    log.Info("Token " + result + " per l'usuari amb id " + userId + " trobat i esborrat.");
                    result = true;
                }
                else
                {
                    log.Info("L'usuari amb id " + userId + " no te token.");
                }
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al esborrar el token d'un usuari.", e);
            }
            log.Debug("deleteUserToken End");
            return result;
        }

        /// <summary>
        /// Esborra la sessió del token
        /// </summary>
        /// <param name="token">Token</param>
        /// <returns>True si ha pogut esborrar la sessio, False si no</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool deleteToken(string token)
        {
            log.Debug("deleteToken Begin");
            bool result = false;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from w in db.WSSessionTokens
                               where w.Token == token
                               select w);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Token " + token + " no existeix.");
                }
                WSSessionToken WSToken = request.First();
                db.WSSessionTokens.Remove(WSToken);
                db.SaveChanges();
                log.Info("Token " + token + " trobat i esborrat.");
                result = true;
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al esborrar un token.", e);
            }
            log.Debug("deleteToken End");
            return result;
        }

        /// <summary>
        /// Actualitza la última data d'accés de la sessió corresponent al token.
        /// Si la sessió ha superat el temps de sessió s'esborra.
        /// </summary>
        /// <param name="token">Token</param>
        /// <returns>True si ha pogut actualitzar la data, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool updateTokenLastAccessDate(string token) 
        {
            log.Debug("updateTokenLastAccessDate Begin");
            cleanOldTokens();
            bool result = false;
            try
            {
                //Seleccionem el primer resultat (es únic).
                var request = (from w in db.WSSessionTokens
                               where w.Token == token
                               select w);
                if (request.Count() != 1)
                {
                    throw new GestorDespesesException("Token " + token + " no existeix.");
                }
                WSSessionToken WSToken = request.First();
                WSToken.lastAccessDateTime = DateTime.Now;
                db.SaveChanges();
                log.Info("Token " + token + " actualitzat");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al actualitzar la data d'accès d'un token.", e);
            }
            log.Debug("updateTokenLastAccessDate End");
            return result;
        }

        /// <summary>
        /// Esborra totes les sessions caducades de la taula de sessions.
        /// </summary>
        /// <returns>True si ha pogut esborrar les sessions caducades, False si no.</returns>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesException">Error en el model controlat.</exception>
        /// <exception cref="GestorDespeses.Model.Exceptions.GestorDespesesUnknownException">Error en el model desconegut.</exception>
        public bool cleanOldTokens()
        {
            log.Debug("deleteUserToken Begin");
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                string ssessionTimeout = coreManager.getSetting(Constants.SESSION_TIMEOUT_KEY);
                // Si no existeix la propietat del sistema per el timeout de session per defecte s'aplica 5 minuts.
                int sessionTimeout = (ssessionTimeout == null) ? 5*60 : Convert.ToInt32(ssessionTimeout);
                DateTime timeLimit = DateTime.Now.AddSeconds(-1*sessionTimeout);
                // Seleccionem totes les sessios que l'ultim acc`s hagi estat fa mes de sessionTimeout segons
                var query = (from w in db.WSSessionTokens where w.lastAccessDateTime < timeLimit select w);
                if (query.Count() > 0)
                {
                    foreach (WSSessionToken token in query)
                    {
                        db.WSSessionTokens.Remove(token);
                    }
                    db.SaveChanges();
                }
                result = true;
                log.Info("Eliminats " + query.Count() + " tokens de la taula de sessions.");
            }
            catch (GestorDespesesException)
            {
                throw;
            }
            catch (Exception e)
            {
                throw new GestorDespesesUnknownException("Error al eliminar els tokens caducats.", e);
            }
            log.Debug("deleteUserToken End");
            return result;
        }
    }
}