﻿using GestorDespeses.BI;
using GestorDespeses.Exceptions;
using GestorDespeses.Model;
using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using System.Web;

namespace GestorDespesesWS.Authorization
{
    public class AuthorizationManager
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        // TODO: Afegir als metodes de isAllowed per sapiguer si l'acces a les dades es de lectura o escriptura.
        public enum AccessType 
        { 
            READ,
            WRITE,
        }

        /// <summary>
        /// Comproba si el token correspon a un usuari del tipus userType
        /// </summary>
        /// <param name="db">Contexte de conexió a la base de dades.</param>
        /// <param name="token">Token</param>
        /// <param name="userType">Tipus d'usuari (WORKER/ADMINISTRATOR)</param>
        /// <returns>True si el token correspon a un usuari de tipus userType. False si no.</returns>
        public static bool isAllowed(GestorDespesesContext db, string token, UserType userType) 
        {
            if (String.IsNullOrEmpty(token))
            {
                log.Error("Token no pot ser una cadena buida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "Token no especificat." });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionManager = new WSSessionBI(db);
                User user = sessionManager.getUserByToken(token);
                result = user.Type.Equals(userType);
            }
            catch(GestorDespesesException)
            {
                log.Error("Token " + token + " no correspon a cap sessió vàlida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_LOGGED, Message = "Token de sessió invàlid." });
            }
            return result;
            
        }

        /// <summary>
        /// Comproba si el token correspon a un usuari am accés a les dades del usuari userId
        /// </summary>
        /// <param name="db">Contexte de conexió a la base de dades.</param>
        /// <param name="token">Token</param>
        /// <param name="userId">Identificador d'usuari</param>
        /// <returns>True si el token correspon a un usuari amb acces a les dades de userId. False si no.</returns>
        public static bool isAllowed(GestorDespesesContext db, string token, int userId)
        {
            if (String.IsNullOrEmpty(token))
            {
                log.Error("Token no pot ser una cadena buida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "Token no especificat." });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionManager = new WSSessionBI(db);
                User user = sessionManager.getUserByToken(token);
                if (user.UserId == userId)
                {
                    // L'usuari té accés al seu propi full de despeses
                    result = true;
                }
                else
                {
                    var query = (from u in user.Workers where u.UserId == userId select u);
                    //Si l'usuari te userId a la llista de treballadors aquest te accès.
                    result = query.Count() == 1;
                }
            }
            catch (GestorDespesesException)
            {
                log.Error("Token " + token + " no correspon a cap sessió vàlida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_LOGGED, Message = "Token de sessió invàlid." });
            }

            return result;
        }

        /// <summary>
        /// Comproba si el token correspon a un usuari amb accés a les dades del full de despeses
        /// </summary>
        /// <param name="db">Contexte de conexió a la base de dades.</param>
        /// <param name="token">Token</param>
        /// <param name="userId">Identificador del full de despeses</param>
        /// <returns>True si el token correspon a un usuari amb acces al full de despeses. False si no.</returns>
        public static bool isAllowedExpenseSheet(GestorDespesesContext db, string token, int expenseSheetId)
        {
            if (String.IsNullOrEmpty(token))
            {
                log.Error("Token no pot ser una cadena buida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "Token no especificat." });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionManager = new WSSessionBI(db);
                User user = sessionManager.getUserByToken(token);
                // Seleccionem el full de despeses
                try
                {
                    ExpensesBI expensesManager = new ExpensesBI(db);
                    ExpensesSheet sheet = expensesManager.getExpensesSheet(expenseSheetId);
                    if (user.UserId == sheet.User.UserId)
                    {
                        // L'usuari té accés al seu propi full de despeses.
                        result = true;
                    }
                    else
                    {
                        var query = (from u in user.Workers where u.UserId == sheet.User.UserId select u);
                        //Si l'usuari te userId a la llista de treballadors aquest te accès.
                        result = query.Count() == 1;
                    }
                }
                catch (GestorDespesesException)
                {
                    // No s'ha trobat el full de despeses, pero retornem false per no donar informació relativa a si un full de despeses existex o no.
                    result = false;
                }
            }
            catch (GestorDespesesException)
            {
                log.Error("Token " + token + " no correspon a cap sessió vàlida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_LOGGED, Message = "Token de sessió invàlid." });
            }

            if (result)
            {
                log.Info("Accés al full de despeses permés.");
            }
            else
            {
                log.Info("Accés al full de despeses denegat.");
            }

            return result;
        }

        /// <summary>
        /// Comproba si el token correspon a un usuari am accés a les dades de la despesa
        /// </summary>
        /// <param name="db">Contexte de conexió a la base de dades.</param>
        /// <param name="token">Token</param>
        /// <param name="userId">Identificador de la despesa</param>
        /// <returns>True si el token correspon a un usuari amb acces a la despesa. False si no.</returns>
        public static bool isAllowedExpense(GestorDespesesContext db, string token, int expenseId)
        {
            if (String.IsNullOrEmpty(token))
            {
                log.Error("Token no pot ser una cadena buida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "Token no especificat." });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionManager = new WSSessionBI(db);
                User user = sessionManager.getUserByToken(token);
                // Seleccionem la despesa
                try
                {
                    ExpensesBI expensesManager = new ExpensesBI(db);
                    Expense expense = expensesManager.getExpense(expenseId);
                    if (expense != null)
                    {
                        if (user.UserId == expense.ExpensesSheet.User.UserId)
                        {
                            // L'usuari té accés al seu propi full de despeses.
                            result = true;
                        }
                        else
                        {
                            var query = (from u in user.Workers where u.UserId == expense.ExpensesSheet.User.UserId select u);
                            //Si l'usuari te userId a la llista de treballadors aquest te accès.
                            result = query.Count() == 1;
                        }
                    }
                }
                catch (GestorDespesesException)
                {
                    // No s'ha trobat la despesa, pero retornem false per no donar informació relativa a si una despesea existex o no.
                    result = false;
                }
            }
            catch (GestorDespesesException)
            {
                log.Error("Token " + token + " no correspon a cap sessió vàlida.");
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_LOGGED, Message = "Token de sessió invàlid." });
            }

            if (result)
            {
                log.Info("Accés a la despesa permés.");
            }
            else
            {
                log.Info("Accés a la despesa denegat.");
            }

            return result;
        }
    }
}