﻿using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.ServiceModel.Web;
using System.Text;
using GestorDespesesWS.Authorization;
using GestorDespeses.Model;
using GestorDespeses.BI;
using GestorDespeses.Exceptions;
using GestorDespeses;


[assembly: log4net.Config.XmlConfigurator(Watch = true)]

namespace GestorDespesesWS
{ 
    // NOTE: You can use the "Rename" command on the "Refactor" menu to change the class name "Service1" in code, svc and config file together.
    // NOTE: In order to launch WCF Test Client for testing this service, please select Service1.svc or Service1.svc.cs at the Solution Explorer and start debugging.
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.PerCall)]
    public class GestorDespesesWS : IGestorDespesesWS
    {
        private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        #region private

        private UserData UserToUserData(User user)
        { 
            //Extraiem els identificadors dels fulls de despeses
            List<int> sheets = new List<int>();
            sheets.AddRange(user.ExpensesSheets.Select(n => n.ExpensesSheetId));

            //Extraiem els identificadors dels usuaris a carrec
            List<int> workers = new List<int>();
            workers.AddRange(user.Workers.Select(n => n.UserId));

            //Creem un nou objecte UserData amb els parametres del objecte User i el retormem
            return new UserData
            {
                UserHandler = user.UserId,
                Username = user.Username,
                Email = user.Email,
                FullName = user.FullName,
                Type = user.Type,
                HeadHandler = user.UserHeadId != null ? (int)user.UserHeadId : -1,
                ExpensesSheetHandlers = sheets,
                WorkerHandlers = workers
            };
        }

        private ExpenseTypeData ExpenseTypeToExpenseTypeData(ExpenseType expenseType) 
        {
            return new ExpenseTypeData
            {
                ExpenseTypeHandler = expenseType.ExpenseTypeId,
                Concept = expenseType.Concept,
                Type = expenseType.Type,
                Value = expenseType.Value,
                MaxValue = expenseType.MaxValue,
                Threshold = expenseType.InvoiceThreshold
            };
        }

        private DayData DayToDayData(Day day)
        {
            return new DayData{
                DayHandler = day.DayId,
                Date = day.Date,
                Type = day.Type
            };
        }

        private ExpenseData ExpenseToExpenseData(Expense expense)
        {
            return new ExpenseData
            {
                ExpenseHandler = expense.ExpensesId,
                Value = expense.Value,
                Description = expense.Description,
                Date = expense.Date,
                Type = expense.Type,
                TypeConcept = expense.TypeConcept,
                TypeValue = expense.TypeValue,
                TypeMaxValue = expense.TypeMaxValue,
                TypeInvoiceThreshold = expense.TypeInvoiceThreshold,
                DateIncrement = expense.DateIncrement,
                IsOnHolidays = expense.IsOnHolidays,
                IsOnWeekend = GestorDespeses.Utils.DateUtils.isDateWeekend(expense.Date),
                HasInvoice = (expense.InvoiceImage != null && expense.InvoiceImage.Length > 0),
                TotalValue = ExpensesBI.ExpenseTotalValue(expense)
            };
        }

        private ExpensesSheetData ExpensesSheetToExpensesSheetData(ExpensesSheet expensesSheet)
        {
            decimal expenseTotal = 0;
            foreach(Expense expense in expensesSheet.Expense)
            {
                expenseTotal += ExpensesBI.ExpenseTotalValue(expense);
            }

            return new ExpensesSheetData
            {
                ExpensesSheetHandler = expensesSheet.ExpensesSheetId,
                Date = expensesSheet.LastStateChangedDate,
                RevokedReason = expensesSheet.RevokedReason,
                State = expensesSheet.State,
                OwnerUsername = expensesSheet.User.Username,
                Value = expenseTotal
            };
        }

        #endregion

        #region Session Control
        public string login(string username, string password)
        {
            string result = null;
            try
            {
                if (String.IsNullOrEmpty(username) || String.IsNullOrEmpty(password))
                {
                    throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "Els arguments 'username' i 'password' no poden estar buits." });
                }
                GestorDespesesContext db = new GestorDespesesContext();
                UserBI usersManager = new UserBI(db);
                WSSessionBI sessionsManager = new WSSessionBI(db);

                if (usersManager.validateUser(username, password))
                {
                    result = sessionsManager.createTokenForUser(usersManager.getUserByUsername(username).UserId);
                }
                else
                {
                    throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.LOGIN_FAILED, Message = "Error al validar l'usuari " + username });
                }
            }
            catch (FaultException fe)
            {
                log.Error(fe.Message);
                throw;
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool logout(string token)
        {
            bool result = false;
            try
            {
                GestorDespesesContext db = new GestorDespesesContext();
                WSSessionBI sessionsManager = new WSSessionBI(db);
                result = sessionsManager.deleteToken(token);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }
            return result;
        }

        #endregion

        #region User Managament
        public List<UserData> listUsers(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            List<UserData> result = new List<UserData>();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            try
            {
                UserBI usersManager = new UserBI(db);
                List<User> users = usersManager.listUsers();
                foreach (User user in users)
                {
                    result.Add(UserToUserData(user));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public List<UserData> listUsersByQuery(string q, int limit, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            List<UserData> result = new List<UserData>();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            try
            {
                UserBI usersManager = new UserBI(db);
                List<User> users = usersManager.listUsersByQuery(q, limit);
                foreach (User user in users)
                {
                    result.Add(UserToUserData(user));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public int addUser(string username,
            string password,
            string email,
            string fullName,
            UserType type,
            string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            int result = -1;
            
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            try
            {
                UserBI usersManager = new UserBI(db);
                result = usersManager.addUser(username,
                                        password,
                                        email,
                                        fullName,
                                        type);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool editUser(int userHandler,
            string password,
            string email,
            string fullName,
            UserType type,
            string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            bool result = false;
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            { 
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat"});
            }
            try
            {
                UserBI usersManager = new UserBI(db);
                result = usersManager.editUser(userHandler,
                                        password,
                                        email,
                                        fullName,
                                        type);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public UserData getUser(int userHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            { 
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat"});
            }
            UserData result = null;
            try
            {
                UserBI usersManager = new UserBI(db);
                User user = usersManager.getUser(userHandler);
                if (user != null)
                {
                    result = UserToUserData(user);
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool deleteUser(int userHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            { 
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat"});
            }
            bool result = false;
            try
            {
                UserBI usersManager = new UserBI(db);
                result = usersManager.deleteUser(userHandler);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool generateValidationCode(string usernameOrEmail) 
        {
            GestorDespesesContext db = new GestorDespesesContext();

            bool result = false;
            try
            {
                UserBI usersManager = new UserBI(db);
                result = usersManager.generateValidationCode(usernameOrEmail);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool changePassword(string newPassword, string activationCode)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            bool result = false;
            try
            {
                UserBI usersManager = new UserBI(db);
                result = usersManager.changePasswordByValidationCode(newPassword, activationCode);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public string generatePassword(int userHandler, string token) 
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            { 
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat"});
            }
            string result = null;
            try
            {
                UserBI usersManager = new UserBI(db);
                result = usersManager.resetPassword(userHandler);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool assignHead(int userHandler, int headHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            { 
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat"});
            }
            bool result = false;
            try 
            {
                UserBI usersManager = new UserBI(db);
                result = usersManager.assignHead(userHandler, headHandler); 
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        #endregion

        #region ExpensesTypes

        public List<ExpenseTypeData> listExpenseTypes(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            { 
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat"});
            }
            List<ExpenseTypeData> res = new List<ExpenseTypeData>();
            try
            {
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                List<ExpenseType> expenseTypes = expenseTypesManager.listExpenseTypes();
                foreach (ExpenseType expenseType in expenseTypes)
                {
                    res.Add(ExpenseTypeToExpenseTypeData(expenseType));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return res;
        }

        public List<ExpenseTypeData> listExpenseTypesByQuery(string q, int limit, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            { 
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat"});
            }
            List<ExpenseTypeData> res = new List<ExpenseTypeData>();

            try
            {
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                List<ExpenseType> expenseTypes = expenseTypesManager.listExpenseTypesByQuery(q, limit);
                foreach (ExpenseType expenseType in expenseTypes)
                {
                    res.Add(ExpenseTypeToExpenseTypeData(expenseType));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return res;
        }

        public int addExpenseType(string concept, decimal? value, decimal? maxValue, decimal? threshold, ExpenseTypeType type, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            int result = -1;
            try
            {
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                result = expenseTypesManager.addExpenseType(concept, value, maxValue, threshold, type);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool editExpenseType(int expenseTypeHandler, string concept, decimal? value, decimal? maxValue, decimal? threshold, ExpenseTypeType type, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                result = expenseTypesManager.editExpenseType(expenseTypeHandler, concept, value, maxValue, threshold, type);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool deleteExpenseType(int expenseTypeHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                result = expenseTypesManager.deleteExpenseType(expenseTypeHandler);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public ExpenseTypeData getExpenseType(int expenseTypeHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            ExpenseTypeData result = null;
            try
            {
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                result = ExpenseTypeToExpenseTypeData(expenseTypesManager.getExpenseType(expenseTypeHandler));
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        #endregion

        #region SystemProperties

        public bool configureSMTP(string fromEmail,
            string fromFullName,
            string password,
            string host,
            int port,
            bool enableSsl,
            string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.editSMTPSettings(fromEmail, fromFullName, password, host, port, enableSsl);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public SMTPData getSMTPSettings(string token) 
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            SMTPData result = new SMTPData();
            try
            {
                CoreBI coreManager = new CoreBI(db);
                try
                {
                    result.host = coreManager.getSetting(Constants.SMTP_HOST_KEY);
                }
                catch (GestorDespesesException) { }
                try
                {
                    result.port = Convert.ToInt32(coreManager.getSetting(Constants.SMTP_PORT_KEY));
                }
                catch (GestorDespesesException) { }
                try
                {
                    result.fromEmail = coreManager.getSetting(Constants.SMTP_FROM_KEY);
                }
                catch (GestorDespesesException) { }
                try
                {
                    result.fromFullName = coreManager.getSetting(Constants.SMTP_USERNAME_KEY);
                }
                catch (GestorDespesesException) { }
                try
                {
                    result.password = coreManager.getSetting(Constants.SMTP_PASSWORD_KEY);
                }
                catch (GestorDespesesException) { }
                try
                {
                    result.enableSsl = Convert.ToBoolean(coreManager.getSetting(Constants.SMTP_SSL_KEY));
                }
                catch (GestorDespesesException) { }
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool setHolidayIncrement(int percentage, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            if (percentage > 100 || percentage < 0)
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "L'argument 'percentage' ha de ser un valor entre 0 i 100" });
            }
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.setSetting(Constants.HOLIDAY_INCREMENT_KEY, Convert.ToString(percentage));
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool setWeekendIncrement(int percentage, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            } 
            if (percentage > 100 || percentage < 0)
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "L'argument 'percentage' ha de ser un valor entre 0 i 100" });
            }
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.setSetting(Constants.WEEKEND_INCREMENT_KEY, Convert.ToString(percentage));
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool setValidationThreshold(int daysToSubstract, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            if (daysToSubstract > 25 || daysToSubstract < 0)
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "L'argument 'daysToSubstract' ha de ser un valor entre 0 i 24" });
            }
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.setSetting(Constants.VALIDATION_THRESHOLD_KEY, Convert.ToString(daysToSubstract));
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }


        public bool setServiceSessionTimeout(int seconds, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            } 
            if (seconds < 60) //Sessio mínima d'un minut
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "L'argument 'seconds' ha de ser un valor superior a 60" });
            }
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.setSetting(Constants.SESSION_TIMEOUT_KEY, Convert.ToString(seconds));
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        #endregion

        #region Calendar

        public List<DayData> listMonthCalendar(int year, int month, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            } 
            List<DayData> res = new List<DayData>();
            try
            {
                CoreBI coreManager = new CoreBI(db);
                List<Day> days = coreManager.listCalendar(year, month);
                foreach (Day day in days)
                {
                    res.Add(DayToDayData(day));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return res;
        }

        public bool setCalendarHoliday(DateTime dateTime, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            } 
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.setCalendarHoliday(dateTime);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool deleteCalendarHoliday(DateTime dateTime, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            } 
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.deleteCalendarHoliday(dateTime);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool setCalendarValidationLimit(DateTime dateTime, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            if (dateTime.Date < DateTime.Now.Date)
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.ARGUMENTS_ERROR, Message = "L'argument 'dateTime' ha de contindre una data superior al moment actual." });
            }
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                result = coreManager.setCalendarValidationLimit(dateTime);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool deleteCalendarValidationLimit(DateTime dateTime, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                CoreBI coreManager = new CoreBI(db);
                Day currenValidationDate = coreManager.getCalendarValidationLimit(dateTime.Year, dateTime.Month);
                if (currenValidationDate.Date.Day == dateTime.Day)
                {
                    coreManager.setDefaultValidationLimit(dateTime.Year, dateTime.Month);
                }
                result = true;
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }
        #endregion

        #region Payment

        public bool payExpensesSheet(int expensesSheetHandler, string token) 
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                ExpensesBI expensesManager = new ExpensesBI(db);
                result = expensesManager.changeExpensesSheetState(expensesSheetHandler, ExpensesSheetState.PAYED, null);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }


        public List<ExpensesSheetData> getExpensesSheetsToPay(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            List<ExpensesSheetData> result = new List<ExpensesSheetData>();
            try
            {
                ExpensesBI expensesManager = new ExpensesBI(db);
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheets(new List<ExpensesSheetState>() { ExpensesSheetState.VALIDATED });
                foreach (ExpensesSheet expensesSheet in sheets)
                {
                    result.Add(ExpensesSheetToExpensesSheetData(expensesSheet));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        #endregion

        #region Worker

        public UserData getCurrentUser(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            UserData result = null;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                User user = sessionsManager.getUserByToken(token);
                if (user != null)
                {
                    result = UserToUserData(user);
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public String getCurrentExpensesRevokedReason(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            string result = "";
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheetOfUser(user.UserId, new List<ExpensesSheetState>() { ExpensesSheetState.OPEN, ExpensesSheetState.REVOKED, ExpensesSheetState.VALIDATING });
                if (sheets.Count == 1)
                {
                    result = sheets.First().RevokedReason;
                }
                else
                {
                    //TODO: llençar excepció que no hi ha full de despeses actiu.¿?¿?
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }
            return result;
        }


        public ExpensesSheetState getCurrentExpensesState(string token) 
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            ExpensesSheetState result = ExpensesSheetState.OPEN;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheetOfUser(user.UserId, new List<ExpensesSheetState>() { ExpensesSheetState.OPEN,  ExpensesSheetState.REVOKED, ExpensesSheetState.VALIDATING });
                if (sheets.Count == 1)
                {
                    result = sheets.First().State;
                }
                else
                {
                    //TODO: llençar excepció que no hi ha full de despeses actiu.¿?¿?
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }
            return result;
        }

        public List<ExpenseData> getCurrentExpenses(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            } 
            List<ExpenseData> result = new List<ExpenseData>();
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheetOfUser(user.UserId, new List<ExpensesSheetState>() { ExpensesSheetState.OPEN, ExpensesSheetState.REVOKED, ExpensesSheetState.VALIDATING });
                if (sheets.Count == 1)
                {
                    List<Expense> expenses = expensesManager.listExpensesOfExpensesSheet(sheets.First().ExpensesSheetId);
                    foreach (Expense expense in expenses)
                    {
                        result.Add(ExpenseToExpenseData(expense));
                    }
                }
                else
                {
                    //TODO: llençar excepció que no hi ha full de despeses actiu.¿?¿?
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }
        
        public bool preValidateCurrentExpenses(string token) 
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheetOfUser(user.UserId, new List<ExpensesSheetState>() { ExpensesSheetState.OPEN });
                if (sheets.Count == 1)
                {
                    result = expensesManager.changeExpensesSheetState(sheets.First().ExpensesSheetId, ExpensesSheetState.VALIDATING, null);
                }
                else
                {
                    //TODO: llençar excepció que no hi ha full de despeses obert.¿?¿?
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool reOpenCurrentExpenses(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheetOfUser(user.UserId, new List<ExpensesSheetState>() { ExpensesSheetState.REVOKED });
                if (sheets.Count == 1)
                {
                    result = expensesManager.changeExpensesSheetState(sheets.First().ExpensesSheetId, ExpensesSheetState.OPEN, null);
                }
                else
                {
                    //TODO: llençar excepció que no hi ha full de despeses obert.¿?¿?
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public List<ExpensesSheetData> getExpenseSheetsHistory(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            List<ExpensesSheetData> result = new List<ExpensesSheetData>();
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheetOfUser(user.UserId, new List<ExpensesSheetState>() { ExpensesSheetState.OPEN, ExpensesSheetState.PAYED, ExpensesSheetState.REVOKED, ExpensesSheetState.VALIDATED, ExpensesSheetState.VALIDATING });
                foreach (ExpensesSheet expensesSheet in sheets)
                {
                    result.Add(ExpensesSheetToExpensesSheetData(expensesSheet));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public int addExpense(decimal value, DateTime date, string description, byte[] image, string expenseTypeConcept, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            int result = -1;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                int expenseTypeId = expenseTypesManager.getExpenseTypeByConcept(expenseTypeConcept).ExpenseTypeId; 
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                result = expensesManager.addExpense(user.UserId, value, date, description, image, expenseTypeId);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public ExpenseData getExpense(int expenseHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            ExpenseData result = null;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                Expense expense = expensesManager.getExpense(expenseHandler);
                if (expense != null)
                {
                    result = ExpenseToExpenseData(expense);
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public byte[] getExpenseInvoice(int expenseHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            byte[] result = null;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                Expense expense = expensesManager.getExpense(expenseHandler);
                if (expense != null)
                {
                    result = expense.InvoiceImage;
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool editExpense(int expenseHandler, decimal value, DateTime date, string description, byte[] image, string expenseTypeConcept, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                ExpensesTypesBI expenseTypesManager = new ExpensesTypesBI(db);
                int expenseTypeId = expenseTypesManager.getExpenseTypeByConcept(expenseTypeConcept).ExpenseTypeId; 
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                result = expensesManager.editExpense(user.UserId, expenseHandler, value, date, description, image, expenseTypeId);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool deleteExpense(int expenseHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowedExpense(db, token, expenseHandler))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                //user no pot ser null, sino el mòdul d'autorització no ens hagues deixat pasar.
                result = expensesManager.deleteExpense(expenseHandler);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public byte[] getInvoice(int expenseHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowedExpense(db, token, expenseHandler))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            byte[] result = null;
            try
            {
                ExpensesBI expensesManager = new ExpensesBI(db);
                result = expensesManager.getExpenseInvoice(expenseHandler);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        #endregion

        #region Head & Workers
        public List<ExpenseData> getExpensesOfExpensesSheet(int expensesSheetHandler, string token)
        {
            // Si el sheet es del propi usuari estem consultan el history
            // Si el sheet no és del propi usuari es el Responsable que l'està validant, per tant el sheet ha d'estar en estat PerValidar.
            GestorDespesesContext db = new GestorDespesesContext();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowedExpenseSheet(db, token, expensesSheetHandler))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            List<ExpenseData> result = new List<ExpenseData>();
            try
            {
                ExpensesBI expensesManager = new ExpensesBI(db);
                List<Expense> expenses = expensesManager.listExpensesOfExpensesSheet(expensesSheetHandler);
                foreach (Expense expense in expenses)
                {
                    result.Add(ExpenseToExpenseData(expense));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }
        #endregion

        #region Head
        public bool isCurrentUserHeadUser(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                //Un usuari serà responsable si té usuaris al seu càrrec.
                WSSessionBI sessionsManager = new WSSessionBI(db);
                User user = sessionsManager.getUserByToken(token);
                if (user.Workers != null && user.Workers.Count > 0) 
                {
                    result = true;
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }
            return result;
        }

        public List<ExpensesSheetData> getExpensesSheetsToValidate(string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();

            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            List<ExpensesSheetData> result = new List<ExpensesSheetData>();
            try
            {
                WSSessionBI sessionsManager = new WSSessionBI(db);
                ExpensesBI expensesManager = new ExpensesBI(db);
                User user = sessionsManager.getUserByToken(token);
                List<ExpensesSheet> sheets = expensesManager.listExpensesSheetToValidate(user.UserId, new List<ExpensesSheetState>() { ExpensesSheetState.VALIDATING });
                foreach (ExpensesSheet expensesSheet in sheets)
                {
                    result.Add(ExpensesSheetToExpensesSheetData(expensesSheet));
                }
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool validateExpensesSheet(int expensesSheetHandler, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER) &&
                !AuthorizationManager.isAllowedExpenseSheet(db,token,expensesSheetHandler))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                ExpensesBI expensesManager = new ExpensesBI(db);
                result = expensesManager.changeExpensesSheetState(expensesSheetHandler, ExpensesSheetState.VALIDATED, null);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }

        public bool revokeExpensesSheet(int expensesSheetHandler, string reason, string token)
        {
            GestorDespesesContext db = new GestorDespesesContext();
            if (!AuthorizationManager.isAllowed(db, token, UserType.ADMINISTRATOR) &&
                !AuthorizationManager.isAllowed(db, token, UserType.WORKER) &&
                !AuthorizationManager.isAllowedExpenseSheet(db, token, expensesSheetHandler))
            {
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.NOT_AUTHORIZED, Message = "Accés no autoritzat" });
            }
            bool result = false;
            try
            {
                ExpensesBI expensesManager = new ExpensesBI(db);
                result = expensesManager.changeExpensesSheetState(expensesSheetHandler, ExpensesSheetState.REVOKED, reason);
            }
            catch (GestorDespesesException gde)
            {
                log.Error(gde.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.BUSINESS_ERROR, Message = gde.Message });
            }
            catch (Exception e)
            {
                log.Error(e.Message);
                throw new FaultException<GestorDespesesFault>(new GestorDespesesFault() { ErrorCode = FaultCode.UNKNOWN_ERROR, Message = "Error desconegut" });
            }

            return result;
        }


        #endregion
    }
}
